# noinspection SpellCheckingInspection
_KUSTO_KEYWORDS = frozenset([
    'abs',
    'accumulate',
    'acos',
    'ago',
    'and',
    'anomalychart',
    'anomalycolumns',
    'areachart',
    'array_concat',
    'array_iif',
    'array_index_of',
    'array_length',
    'array_rotate_left',
    'array_rotate_right',
    'array_shift_left',
    'array_shift_right',
    'array_slice',
    'array_split',
    'avg',
    'avgif',
    'axes',
    'bag_keys',
    'barchart',
    'bin',
    'bin_at',
    'bin_auto',
    'card',
    'ceiling',
    'columnchart',
    'contains',
    'cos',
    'countif',
    'dcount',
    'dcount_hll',
    'dcountif',
    'dynamic',
    'endofday',
    'endofmonth',
    'endofweek',
    'endofyear',
    'endswith',
    'exp',
    'false',
    'FALSE',
    'False',
    'floor',
    'format_datetime',
    'format_timespan',
    'getmonth',
    'gettype',
    'getyear',
    'hash',
    'hourofday',
    'iff',
    'iif',
    'in',
    'isempty',
    'isfinite',
    'isinf',
    'isnan',
    'isnotempty',
    'isnotnull',
    'isnull',
    'kind',
    'ladderchart',
    'legend',
    'linechart',
    'log',
    'loggamma',
    'make_bag',
    'make_datetime',
    'make_list',
    'make_set',
    'matches',
    'max',
    'maxif',
    'min',
    'minif',
    'none',
    'not',
    'now',
    'null',
    'or',
    'pack',
    'pack_array',
    'panels',
    'parse_json',
    'percentiles',
    'piechart',
    'pivotchart',
    'pow',
    'regex',
    'round',
    'scatterchart',
    'series',
    'set_difference',
    'set_has_element',
    'set_intersect',
    'set_union',
    'sign',
    'split',
    'sqrt',
    'stacked',
    'stacked100',
    'stackedareachart',
    'startofday',
    'startofmonth',
    'startofweek',
    'startofyear',
    'startswith',
    'stdev',
    'stdevif',
    'stdevp',
    'strcat_array',
    'strcat_delim',
    'strcmp',
    'string_size',
    'strlen',
    'strrep',
    'substring',
    'sum',
    'sumif',
    'table',
    'timechart',
    'timepivot',
    'title',
    'tobool',
    'todatetime',
    'tohex',
    'toint',
    'tolong',
    'tolower',
    'tostring',
    'toupper',
    'true',
    'TRUE',
    'True',
    'unstacked',
    'variance',
    'varianceif',
    'variancep',
    'xaxis',
    'xcolumn',
    'xtitle',
    'yaxis',
    'ycolumns',
    'ymax',
    'ymin',
    'ysplit',
    'ysplit',
    'ytitle',
])
